mod cpu;
mod parser;

use std::env::args;
use std::fs::read_to_string;

use cpu::*;

use parser::{parse_program, ParseError};

// entry point of our basic CLI
pub fn main() {
    let args: Vec<_> = args().collect();

    if args.len() < 2 {
        println!("Forgot the input file?");
    } else {
        match read_to_string(args[1].clone()) {
            Ok(s) => match parse_program(&s[..]) {
                Ok(prog) => {
                    let mut processor = CPU::load(&prog);
                    for i in 2..(args.len()) {
                        match args[i].clone().parse() {
                            Ok(n) => processor.add_input(n),
                            Err(_) => println!("Warning: {} is not a number", args[i]),
                        }
                    }

                    println!("program loaded!");
                    println!("Starting memory:\n");
                    showstate(&processor);
                    println!("Executing...\n");

                    match processor.exec() {
                        Err(CpuError::BadRegister(l, r)) => {
                            println!(
                                "Error, register {} does not exist! Program counter at: {}",
                                r, l
                            )
                        }
                        Err(CpuError::OverFlow(l)) => {
                            println!("Error, Overflow! Program counter at: {}", l)
                        }
                        Err(CpuError::EndOfInput(l)) => {
                            println!("Error, end of input! Did you forget to add enough input numbers? Program counter at {}", l)
                        }
                        Err(CpuError::BadJump(l)) => {
                            println!("Error, bad jump instruction! Program counter was {}", l)
                        }
                        Err(CpuError::AccAccess(l)) => {
                            println!(
                                "Error, tried to access the accumulator! Program counter at {}",
                                l
                            )
                        }

                        Err(CpuError::ProgramIsOver(l)) => {
                            println!(
                                "Error, program is over! Did you forget to HALT? Program counter at {}",
                                l
                            )
                        }
                        Ok(()) => {}
                    }

                    println!("Done. Final memory:\n");

                    showstate(&processor);
                }
                Err(ParseError::ErrorAtLine(l)) => {
                    println!(
                        "There was an error parsing the file! Dude! Check line {}",
                        l + 1
                    )
                }
            },
            Err(_) => println!("Cannot read file {}!", args[1]),
        }
    }
}

fn showmem(cpu: &CPU) {
    print!("Memory: ");
    let (k, _, _, _) = cpu.exposestate();
    for i in 0..k.len() - 1 {
        print!(" {} |", k[i])
    }
    println!(" {}", k[k.len() - 1])
}

fn showinput(cpu: &CPU) {
    print!("Input: ");
    let (_, k, _, _) = cpu.exposestate();
    for i in 0..k.len() {
        print!(" {} ", k[i])
    }
    print!("\n");
}

fn showoutput(cpu: &CPU) {
    print!("Output: ");
    let (_, _, k, _) = cpu.exposestate();
    for i in 0..k.len() {
        print!(" {} ", k[i])
    }
    print!("\n");
}

fn showstate(cpu: &CPU) {
    showmem(cpu);
    showinput(cpu);
    showoutput(cpu);
}
