use crate::cpu::{Instruction, RegRef};
use regex::{Regex, RegexSet};
use std::collections::VecDeque;
use Instruction::*;
use RegRef::*;

// OOO WE PARSIN' WITH REGEX!!!

#[derive(Debug, PartialEq, Eq)]
pub enum ParseError {
    ErrorAtLine(u32), // we only return the offending line.
}

// a few helper function to delete comments and newlines, while keeping the line number
// consistent

fn count_initial_lines(prog: &str) -> u32 {
    return prog
        .split("\n")
        .map(|x| x.trim())
        .take_while(|&x| x == "")
        .count() as u32;
}

pub fn parse_program(prog: &str) -> Result<Vec<Instruction>, ParseError> {
    let s = prog.trim();

    let mut posy: u32 = count_initial_lines(prog); // hoping no one will write a program this long...

    let mut res: VecDeque<Instruction> = vec![].into();

    let lines = s.split("\n");

    for linefull in lines {
        let line = linefull.trim();

        // ignore comments
        if line.starts_with("#") {
            posy += 1;
            continue;
        }

        // skip blank lines
        if line.is_empty() {
            posy += 1;
            continue;
        }

        match parse_line(line) {
            Some(instr) => {
                res.push_back(instr);

                posy += 1;
            }
            None => return Err(ParseError::ErrorAtLine(posy)),
        }
    }

    return Ok(res.into());
}

// where the MAGIC happens!

// first we define a container for closures we're gonna use later.
// we'll define a list of pairs, where the first element of the pair
// is a closure that takes a string and builds the actual instruction
// and the second element is a regex matching the said instruction.
#[derive(Clone, Copy)]
pub struct FnWrap {
    pub f: fn(&str) -> Instruction,
}

impl FnWrap {
    fn new(f: fn(&str) -> Instruction) -> Self {
        FnWrap { f }
    }
}

pub fn parse_line(line: &str) -> Option<Instruction> {
    let s = line.trim();

    // comping the example from https://docs.rs/regex/latest/regex/struct.RegexSet.html

    // COULD BE BETTER!!!!!!!!!
    // We have to parse differently in case of references..
    let patternspairs = [
        (
            FnWrap::new(|x| READ(Direct(x.parse().unwrap()))),
            r"^READ\s+(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| READ(Indirect(x.parse().unwrap()))),
            r"^READ\s+\*(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| WRITE(Direct(x.parse().unwrap()))),
            r"^WRITE\s+(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| WRITE(Indirect(x.parse().unwrap()))),
            r"^WRITE\s+\*(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| LOAD(Direct(x.parse().unwrap()))),
            r"^LOAD\s+(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| LOAD(Indirect(x.parse().unwrap()))),
            r"^LOAD\s+\*(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| LOADI(x.parse().unwrap())),
            r"^LOAD\s+=\s*([-+]?\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| STORE(Direct(x.parse().unwrap()))),
            r"^STORE\s+(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| STORE(Indirect(x.parse().unwrap()))),
            r"^STORE\s+\*(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| ADD(Direct(x.parse().unwrap()))),
            r"^ADD\s+(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| ADD(Indirect(x.parse().unwrap()))),
            r"^ADD\s+\*(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| ADDI(x.parse().unwrap())),
            r"^ADD\s=\s*([-+]?\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| SUB(Direct(x.parse().unwrap()))),
            r"^SUB\s+(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| SUB(Indirect(x.parse().unwrap()))),
            r"^SUB\s+\*(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| SUBI(x.parse().unwrap())),
            r"^SUB\s=\s*([-+]?\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| JUMP(x.parse().unwrap())),
            r"^JUMP\s+(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| JZERO(x.parse().unwrap())),
            r"^JZERO\s+(\d+)\s*(?:#.*)?$",
        ),
        (
            FnWrap::new(|x| JGTZ(x.parse().unwrap())),
            r"^JGTZ\s+(\d+)\s*(?:#.*)?$",
        ),
        (FnWrap::new(|_| HALT), r"^HALT\s*(?:#.*)?$"),
    ];

    // Compile a set matching any of our patterns.
    let set = RegexSet::new(&patternspairs.map(|(_, y)| y)).unwrap();
    // Compile each pattern independently.
    let regexes: Vec<_> = patternspairs
        .iter()
        .map(|(f, pat)| (f, Regex::new(pat).unwrap()))
        .collect();

    // Match against the whole set first and identify the individual
    // matching patterns.
    let matches: Vec<_> = set.matches(s).into_iter().collect(); // we should only have one match

    if matches.len() < 1 {
        // if there's less than one match, nothing matched! Wohoo!!
        return None;
    }

    // let's pick the appropriate constructor function, and
    let (f, reg) = &regexes[matches[0]];

    // we run the regex that matched again to get the capture group.
    // then we just use the function f to build the instructon
    let param = reg.captures(s).unwrap().get(1).map_or("", |x| x.as_str());

    // finally!
    return Some((f.f)(param));
}

// testing should go in /tests but i am LAZY!!
#[test]
fn single_line_test() {
    assert_eq!(parse_line("READ 3"), Some(READ(Direct(3))));
    assert_eq!(parse_line("READ *3"), Some(READ(Indirect(3))));
    assert_eq!(parse_line("WRITE 3"), Some(WRITE(Direct(3))));
    assert_eq!(parse_line("WRITE *3"), Some(WRITE(Indirect(3))));
    assert_eq!(parse_line("LOAD 3"), Some(LOAD(Direct(3))));
    assert_eq!(parse_line("LOAD *3"), Some(LOAD(Indirect(3))));
    assert_eq!(parse_line("LOAD =3"), Some(LOADI(3)));
    assert_eq!(parse_line("STORE 3"), Some(STORE(Direct(3))));
    assert_eq!(parse_line("STORE *3"), Some(STORE(Indirect(3))));
    assert_eq!(parse_line("ADD 3"), Some(ADD(Direct(3))));
    assert_eq!(parse_line("ADD *3"), Some(ADD(Indirect(3))));
    assert_eq!(parse_line("ADD =3"), Some(ADDI(3)));
    assert_eq!(parse_line("SUB 3"), Some(SUB(Direct(3))));
    assert_eq!(parse_line("SUB *3"), Some(SUB(Indirect(3))));
}

#[test]
fn program_test() {
    let code = "# this is a comment. \nREAD 1 # this is another comment.\nLOAD 1\nJZERO 8\nSUB =-1\nREAD 1\nWRITE *1\nJUMP 3\nHALT";

    let program = vec![
        READ(Direct(1)),
        LOAD(Direct(1)),
        JZERO(8),
        SUBI(-1),
        READ(Direct(1)),
        WRITE(Indirect(1)),
        JUMP(3),
        HALT,
    ];

    assert_eq!(parse_program(code), Ok(program));
}

#[test]
fn countlines_initial_test() {
    assert_eq!(count_initial_lines("  \n \n \nwow!\n\n"), 3)
}

#[test]
fn full_line_count_test() {
    let prog = "
        # This is a test program with bad syntax. 
        # The error should be on line 3.
        JUMP JUMP JUMP
        ";

    let parsed = parse_program(prog);

    assert_eq!(Err(ParseError::ErrorAtLine(3)), parsed)
}
